/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.yuandian.dataflow.statemachine.client;


 
import org.apache.ratis.client.RaftClient;
import org.apache.ratis.conf.Parameters;
import org.apache.ratis.conf.RaftProperties;
import org.apache.ratis.grpc.GrpcConfigKeys;
import org.apache.ratis.grpc.GrpcFactory;
import org.apache.ratis.protocol.ClientId;
import org.apache.ratis.protocol.Message;
import org.apache.ratis.protocol.RaftClientReply;
import org.apache.ratis.protocol.RaftGroup;
import org.apache.ratis.protocol.RaftGroupId;
import org.apache.ratis.protocol.RaftPeer;
import org.apache.ratis.protocol.RaftPeerId;
import org.apache.ratis.protocol.RaftRpcMessage;
import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
import org.apache.ratis.util.NetUtils;
import org.springframework.cglib.proxy.CallbackFilter;

import com.alibaba.nacos.shaded.io.grpc.netty.shaded.io.netty.handler.codec.protobuf.ProtobufEncoder;
import com.yuandian.dataflow.statemachine.StateServer;
import com.yuandian.dataflow.statemachine.Operate.OperateType;
import com.yuandian.dataflow.statemachine.state.WorkerState;
import com.yuandian.dataflow.statemachine.Operate;
import com.yuandian.dataflow.statemachine.Query;

import lombok.extern.slf4j.Slf4j;

import java.io.IOException;
import java.nio.charset.Charset;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Counter client application, this application sends specific number of
 * INCREMENT command to the Counter cluster and at the end sends a GET command
 * and print the result
 * <p>
 * Parameter to this application indicate the number of INCREMENT command, if no
 * parameter found, application use default value which is 10
 */
@Slf4j
public final class CounterClient {

  private CounterClient(){
  }

 
  public static void main(String[] args)
      throws IOException, InterruptedException {
    //indicate the number of INCREMENT command, set 10 if no parameter passed
    int increment = args.length > 0 ? Integer.parseInt(args[0]) : 10;

    //build the counter cluster client
    RaftClient raftClient = buildClient();

    //use a executor service with 10 thread to send INCREMENT commands
    // concurrently
    ExecutorService executorService = Executors.newFixedThreadPool(10);

    increment = 10;
    CountDownLatch latch = new CountDownLatch(increment);
    //send INCREMENT commands concurrently
    System.out.printf("Sending %d increment command...%n", increment);
    Instant now = Instant.now();
    for (int i = 0; i < increment; i++) {
      
      var op = new Operate(OperateType.PUT_WORKERSTATE, new WorkerState(null)); 
      var reply = raftClient.io().send(op);
    
      log.info("{}", reply);
      executorService.submit(() -> raftClient.io().send(op));
          latch.countDown();
    }
    
    

    //shutdown the executor service and wait until they finish their work
    executorService.shutdown();
    executorService.awaitTermination(increment * 500L, TimeUnit.MILLISECONDS);
    latch.await();
    log.info("{}", Duration.between(now, Instant.now()).toMillis());
    //send GET command and print the response
 
    var query = new Query(Query.Type.GET_STATE, new WorkerState(null));
    RaftClientReply count = raftClient.io().sendReadOnly(query);
    String response = count.getMessage().getContent().toString(Charset.defaultCharset());
    // System.out.println(response);
    log.info("{}", response);
    
  }

  /**
   * build the RaftClient instance which is used to communicate to
   * Counter cluster
   *
   * @return the created client of Counter cluster
   */
  private static RaftClient buildClient() {
    RaftProperties raftProperties = new RaftProperties();
        //set the storage directory (different for each peer) in RaftProperty object
    

    //set the port which server listen to in RaftProperty object
    // final int port = NetUtils.createSocketAddr("localhost:4440").getPort();
    // GrpcConfigKeys.Server.setPort(raftProperties, port);

    var peers = new ArrayList<RaftPeer>();
    String[] addresses = new String[]{"localhost:4440","localhost:4441","localhost:4442"};

    for (int i = 0; i < addresses.length; i++) {
      var port = addresses[i].split(":")[1];
      peers.add(RaftPeer.newBuilder().setId("" + port).setAddress(addresses[i]).build());
    }
    RaftGroup raftGroup = RaftGroup.valueOf(
    RaftGroupId.valueOf(StateServer.CLUSTER_GROUP_ID), peers);
    
    RaftClient.Builder builder = RaftClient.newBuilder()
        .setProperties(raftProperties)
        .setRaftGroup(raftGroup)
        .setClientRpc(new GrpcFactory(new Parameters()).newRaftClientRpc(ClientId.randomId(), raftProperties));
    return builder.build();
  }
}
